/*************************************************************************
*
*  Based on DMAW.C  ver 1.01  (Aug 15, 94)
*
*  Copyright (C) 1994-96 Creative Technology.
*
*  DMA DEMO PROGRAM FOR PLAYING WAVE FILES
*
*  PURPOSE:     This program demonstrates how to play a .wav file
*               using DMA auto-init mode.
*
*  LIMITATION : This program does not support 8 bit STEREO for SBPro.
*
*               16 bit files must use the SB16.
*
*  DISCLAIMER : Although this program has been tested with
*               standard 8/16 bit PCM WAVE files, there could
*               exist some unknown bugs.
*
*  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
*  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
*  PURPOSE.
*
*  You have a royalty-free right to use, modify, reproduce and
*  distribute the Sample Files (and/or any modified version) in
*  any way you find useful, provided that you agree that
*  Creative has no warranty obligations or liability for any Samples Files.
*
*
**************************************************************************/

/* 
 * Bug fixes (mainly for weird soundcards with funny 16 bit IRQs) 
 * Port to DJGPP environment and rewritten to give a streaming interface
 * copyright (c) 1997-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Uses the routines in mlib.c for the IRQ and DMA control
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "mlib.h"
#include "jukebox.h"

// Use a number of small buffers
#define DMA_BLK_SIZE	 (31*1024)
#define MAX_BUF			2	// 31*2K=62K
int totalDmaBufSize;


#define DSP_DATA_AVAIL               0xE
#define DSP_READ_PORT                0xA
#define DSP_READY                   0xAA
#define DSP_RESET                    0x6
#define DSP_TIME_CONSTANT         0x0040
#define DSP_WRITE_PORT               0xC
#define DSP_VERSION                  0xE1

#define MIXER_ADDR      0x4
#define MIXER_DATA      0x5

#define SINGLE_CYCLE                0
#define AUTO_INIT                   1

#define SUCCESS                     1
#define FAIL                        0

#define MONO                        0
#define STEREO                      1


/*----------------------------------------------------------------*/
/*---------  GLOBAL DECLARATIONS  --------------------------------*/
/*----------------------------------------------------------------*/

/* Values read from the BLASTER variable, or found initially */
static BYTE  DMAChan8Bit,
      DMAChan16Bit,
      IRQNumber;

static BYTE g16BitDMA;
static WORD sb_Base=0,DSP_Ver;

static dma_memory_block_t dmaBlock;
static irq_info_t IrqInfo;
static int dmaBufInsert;

// Data about what we are trying to play
static int bitsPerSample;
static int sampleRate;
static BOOL is_mono;
static BOOL speakerOn;


/*************************************************************************
*
* FUNCTION: DSPOut()
*
* DESCRIPTION: Writes the value passed to this function to the DSP chip.
*
*************************************************************************/
static void DSPOut(int WriteValue)
{
  // Wait until DSP is ready before writing the command
  while ((inp(sb_Base + DSP_WRITE_PORT) & 0x80) != 0);

  outp(sb_Base + DSP_WRITE_PORT, WriteValue);
  return;
}

/*************************************************************************
*
*  FUNCTION: start_playback()
*
*  DESCRIPTION : Sets up playing of the wave file depending on number
*         of bits per sample, MONO/STEREO and DMAMode
*
*************************************************************************/
static void start_playback(WORD BytesToPlay, BYTE DMAMode)
{ WORD samples,cmd;

  /* Get number of samples (minus one) (stereo pair = single sample) */
  samples=BytesToPlay;
  if (bitsPerSample==16) samples>>=1;
  if (samples>0) samples--;

  if(DSP_Ver < 4) // SBPro (DSP ver 3.xx)
  {
    if(bitsPerSample == 8)
    { 
      if (DMAMode == AUTO_INIT)
      { DSPOut(0x0048);
        DSPOut(samples & 0x00FF);
        DSPOut(samples >> 8);
        DSPOut(0x001C);  // AUTO INIT 8bit PCM
      }
      else
      { DSPOut(0x0014);  // SINGLE CYCLE 8bit PCM
        DSPOut(samples & 0x00FF);
        DSPOut(samples >> 8);
      }
    }
    else if (bitsPerSample == 16) // 16Bit
    { 
      /* Command 0x41 = Output, followed by sample rate */
      DSPOut(0x0041);
      DSPOut((WORD)((sampleRate>>8) & 0xFF));      // Hi byte
      DSPOut((WORD)(sampleRate & 0xFF));           // Lo byte

      /* DSP Command */
      cmd=(g16BitDMA) ? 0x00B0 : 0x00C0;
      if (DMAMode == AUTO_INIT) cmd|=0x0004;
      DSPOut(cmd);
      DSPOut((is_mono) ? 0x0010 : 0x0030);  // MONO/STEREO
      DSPOut(samples & 0x00FF);     // LO byte size
      DSPOut(samples >> 8);         // HI byte size
    }
  }
  else if(DSP_Ver == 4)// SB16 (DSP ver 4.xx)
  {
#if 0
    /* Command 0x41 = Output, followed by sample rate */
    DSPOut(0x0041);
    DSPOut((WORD)((sampleRate>>8) & 0xFF));        // Hi byte
    DSPOut((WORD)(sampleRate & 0xFF));             // Lo byte

    /* DSP Command */
    cmd=(g16BitDMA) ? 0x00B0 : 0x00C0;
    if (DMAMode == AUTO_INIT) cmd|=0x0006;
    DSPOut(cmd);

    /* Mode value */
    cmd=(bitsPerSample==8) ? 0x0000 : 0x0010;
    if (!is_mono) cmd|=0x0020;
    DSPOut(cmd);
    DSPOut(samples & 0x00FF);   // LO byte
    DSPOut(samples >> 8);       // HI byte
#else
    /* Command 0x41 = Output, followed by sample rate */
    DSPOut(0x0041);
    DSPOut((WORD)((sampleRate>>8) & 0xFF));        // Hi byte
    DSPOut((WORD)(sampleRate & 0xFF));             // Lo byte

    cmd=(bitsPerSample==8) ? 0x00C0 : 0x00B0;
    if (DMAMode == AUTO_INIT) cmd|=0x0006;
    DSPOut(cmd);

    /* Mode value */
    cmd=(bitsPerSample==8) ? 0x0000 : 0x0010;
    if (!is_mono) cmd|=0x0020;
    DSPOut(cmd);
    DSPOut(samples & 0x00FF); // LO byte
    DSPOut(samples >> 8);     // HI byte
#endif
  }
}


/*************************************************************************
*
* FUNCTION: InitDMADSP()
*
* DESCRIPTION: This function reads the first data block of the file and
*              from it obtains information that is needed to program the
*              DMA and DSP chips.  After reading the data block, the file
*              pointer points to the first byte of the voice data.
*
*              NOTE: The DMA chip is ALWAYS programmed for auto-init mode
*                    (command 0x58)!  The DSP chip will be programmed for
*                    auto-init or single-cycle mode depending upon
*                    conditions--see start_playback() for details.
*
*************************************************************************/
static char InitDMADSP(void)
{ BYTE ByteTimeConstant;
  BYTE channel;

  channel=(bitsPerSample==8) ? DMAChan8Bit : DMAChan16Bit;
  g16BitDMA=(channel>=4) ? TRUE : FALSE;

  if (!DmaInit(&dmaBlock,channel))
	  return FAIL;

  /*--- PROGRAM THE DSP CHIP ------------------------------------------*/
  /*-------------------------------------------------------------------*/
  if(DSP_Ver < 4)
  {
    ByteTimeConstant = (BYTE) (256 - 1000000/sampleRate);
    DSPOut((int) DSP_TIME_CONSTANT);
    DSPOut((int) ByteTimeConstant);
  }

  return SUCCESS;
}

static void ms_delay(int n_ms)
{ int i;
  for (i=0 ; i<n_ms ; i++)
   { int j;
     for (j=0 ; j<4000 ; j++) inp(0x80);
   }
}

/*************************************************************************
*
* FUNCTION: ResetDSP()
*
* DESCRIPTION: Self explanatory
*
*************************************************************************/
static char ResetDSP()
{
  outp(sb_Base + DSP_RESET, (int) 1);
  ms_delay(10); // wait 10 mS
  outp(sb_Base + DSP_RESET, (int) 0);

  // Wait until data is available
  while ((inp(sb_Base + DSP_DATA_AVAIL) & 0x80) == 0);

  if (inp(sb_Base + DSP_READ_PORT) == DSP_READY)
  {
    outp(sb_Base + DSP_WRITE_PORT, DSP_VERSION);
    while ((inp(sb_Base + DSP_DATA_AVAIL) & 0x80) == 0);
    DSP_Ver = inp(sb_Base + DSP_READ_PORT);
    inp(sb_Base + DSP_READ_PORT);
    return(SUCCESS);
  }

  return(FAIL);
}

/*************************************************************************
*
* FUNCTION: DMAOutputISR()
*
* DESCRIPTION:  Interrupt service routine.  Every time the DSP chip finishes
*               playing half of the DMA buffer in auto-init mode, an
*               interrupt is generated, which invokes this routine.
*
*************************************************************************/
irq_interrupt_t DMAOutputISR(void)
{ WORD target;
  int canCopy;

  if (DSP_Ver == 4)
  { int IntStatus;
    outp(sb_Base + 4, 0x82);       // Select interrupt status reg. in mixer
    IntStatus = inp(sb_Base + 5);  // Read interrupt status reg.

    // acknowledge 16-bit DMA
    if (IntStatus & 2) inp(sb_Base + 0xF);
    // acknowledge 8-bit DMA
    if (IntStatus & 1) inp(sb_Base + 0xE);

  } else
  { // Acknowledge interrupt (8-bit)
    inp(sb_Base + (int) DSP_DATA_AVAIL);  
  }

  // Acknowledge the PC's interrupt controller
  IrqAck(&IrqInfo);
}


/**************************************************************************
*
* FUNCTION: audioPlay(char *buffer,int nbytes)
*
* DESCRIPTION: add samples to the play buffer, return number of bytes copied
**************************************************************************/
int audioPlay(char *buffer,int nBytes)
{ long canCopy,currentPos;

  currentPos=DmaPos(&dmaBlock);
  // Don't bother if there is less than 256 bytes
  if (dmaBufInsert<currentPos && (currentPos-dmaBufInsert)<256) return 0;

  canCopy=(dmaBufInsert>currentPos) ? 
	  (totalDmaBufSize-dmaBufInsert) : (currentPos-dmaBufInsert);
//printf("%l5X %5lX %5lx %5lx\n",dmaBufInsert,currentPos,canCopy,nBytes);
  if (canCopy>nBytes) canCopy=nBytes;
  if (canCopy>0)
  { DmaCopy(&dmaBlock,dmaBufInsert,canCopy,buffer);
    dmaBufInsert=(dmaBufInsert+canCopy)%totalDmaBufSize;
  }

  if (!speakerOn)
  { DSPOut(0x00D1);  // Turn speaker On
    speakerOn=TRUE;
  }

  return canCopy;
}


/**************************************************************************
*
* FUNCTION: audioInit(int initRate,int initBits,int initChans)
*
* DESCRIPTION: Initialise the audio hardware
**************************************************************************/
int audioInit(int initRate,int initBits,int initChans)
{ int err;
  char *EnvString;
  DWORD val;

  sampleRate=initRate;
  bitsPerSample=initBits;
  is_mono=(initChans==1);
  speakerOn=FALSE;

  /*--- SET EVERYTHING TO ZERO TO START WITH ---------------------*/
  /*--------------------------------------------------------------*/
  DMAChan8Bit=DMAChan16Bit=IRQNumber=99;
  sb_Base=DSP_Ver=0;

  /*--- PROCESS THE BLASTER ENVIRONMENT VARIABLE -----------------*/
  /*--------------------------------------------------------------*/
  EnvString = getenv("BLASTER");
  if (EnvString == NULL)
  { fputs("BLASTER environment variable not set !",stderr);
    return 1;
  }

  while (*EnvString != '\0')
   switch(toupper(*EnvString))
    { case 'A':  // I/O base port address found
        val=strtoul(EnvString+1,&EnvString,16);
        if (val<0x100 || val>=0x4000) break;
        sb_Base=(WORD)val;
        break;

      case 'D': // 8-bit DMA channel
        val=strtoul(EnvString+1,&EnvString,10);
        if (val!=0 && val!=1 && val!=3) break;
        DMAChan8Bit=(BYTE)val;
        break;

      case 'H': // 16-bit DMA channel
        val=strtoul(EnvString+1,&EnvString,10);
        if (val>7 || val==2 || val==4) break;
        DMAChan16Bit=(BYTE)val; 
        break;

      case 'I':  // IRQ number
        val=strtoul(EnvString+1,&EnvString,10);
        if (val==2 || val>15) break;
        IRQNumber=(BYTE)val;
        break;

      default:
        EnvString++;
        break;
    }

  /*--- IF SOMETHING IS MISSING PRINT WHAT WE HAVE ---------------*/
  /*--------------------------------------------------------------*/
  if (IRQNumber==99 || sb_Base==99 || (DMAChan8Bit==99 && DMAChan16Bit==99))
  { fprintf(stderr,"BLASTER = %s\n",getenv("BLASTER"));
    fprintf(stderr,"  8-bit DMA channel  = %d\n",DMAChan8Bit);
    fprintf(stderr," 16-bit DMA channel  = %d\n",DMAChan16Bit);
    fprintf(stderr," I/O Port address    = 0x%X\n",sb_Base);
    fprintf(stderr," IRQ Number          = %d\n",IRQNumber);
	sb_Base=0;
    return 2;
  }

  if(ResetDSP() == FAIL)
  {
    fputs("Unable to reset DSP chip !",stderr);
    return 3;
  }

  if((DSP_Ver < 4) && (bitsPerSample == 16))
  { fprintf(stderr,"** DSP version %d.xx does not support 16bit files.\n", DSP_Ver);
    return 4;
  }

  /*--- ALLOCATE BUFFERS -----------------------------------------*/
  /*--------------------------------------------------------------*/
  /* Start with 31K - 1K = 30K, we will never get a 64K buffer */
  totalDmaBufSize=DMA_BLK_SIZE*MAX_BUF;
  while (totalDmaBufSize>0)
  { if (DmaAlloc(&dmaBlock,totalDmaBufSize))
	{ break;
	}
	totalDmaBufSize-=DMA_BLK_SIZE;
  }
  if (totalDmaBufSize==0)
  { fputs("DMA buffer allocation failed\n",stderr);
	return 5;
  }
//  printf("Using %d buffers (%dK)\n",totalDmaBufSize/DMA_BLK_SIZE,totalDmaBufSize/1024);

  // Initialise the FIFO
  dmaBufInsert=DMA_BLK_SIZE;

  /*--- BEGIN PLAYING (silence) --------------------------------------*/
  /*------------------------------------------------------------------*/
  start_playback(DMA_BLK_SIZE, AUTO_INIT);

  IrqInit(&IrqInfo, IRQNumber, DMAOutputISR);

  /*--- PROGRAM THE DMA, DSP CHIPS -----------------------------------*/
  /*------------------------------------------------------------------*/
  if (InitDMADSP() == FAIL)
  { fputs("InitDMADSP() failed\n",stderr);
    IrqEnd(&IrqInfo);
    DmaFree(&dmaBlock);
    return 6;
  }

  return 0;
}


/**************************************************************************
*
* FUNCTION: audioComplete()
*
* DESCRIPTION: Wait for the currently buffered audio data to finish playing
**************************************************************************/
void audioComplete(void)
{ long currentPos,toPlay;

  do { currentPos=DmaPos(&dmaBlock);
	   toPlay=(currentPos-dmaBufInsert+totalDmaBufSize)%totalDmaBufSize;
	 }
  while (toPlay>4410);
}


/**************************************************************************
*
* FUNCTION: end_player()
*
* DESCRIPTION: Opposite of audioInit()
**************************************************************************/
void audioExit(void)
{
  printf("Audio Exiting\n");
  start_playback(256, SINGLE_CYCLE);
	
  if (speakerOn)
  { DSPOut(0x00D3);  // Turn speaker Off
    speakerOn=FALSE;
  }

  // Halt the DMA controller
  DSPOut((g16BitDMA==FALSE) ? 0x00D9 : 0x00DA);  // Done playing, halt DMA

  // First part of ResetDSP
  outp(sb_Base + DSP_RESET, (int) 1);
  delay(10); // wait 10 mS
  outp(sb_Base + DSP_RESET, (int) 0);

  /*--- RESTORE ISR AND ORIGINAL IRQ VECTOR -------------------------*/
  /*-----------------------------------------------------------------*/
  IrqEnd(&IrqInfo);

  // Free the memory
  DmaFree(&dmaBlock);

  // Clear out the settings
  DMAChan8Bit=DMAChan16Bit=IRQNumber=0;
  sb_Base=DSP_Ver=0;
}


/**************************************************************************
*
* FUNCTION: void volume_write(int port,int val)
*
* DESCRIPTION: Set the selected volume control to the desired level, returns true if suceeded
**************************************************************************/
bool volume_write(int port,int val)
{ if (sb_Base==0 || speakerOn==0) return false;
  outp(sb_Base + MIXER_ADDR, port);
  outp(sb_Base + MIXER_DATA, val);
  outp(sb_Base + MIXER_ADDR, port+1);
  outp(sb_Base + MIXER_DATA, val);
  return true;
}
