/* 
 * Jukebox hardware interface
 * -------------------------------------
 * copyright (c) 1997-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Uses the parallel port to connect to the physical hardware
 * Data outputs 0-7 are connected to the button columns (each via a diode)
 * The rows are connected to ERROR, SELECT, PAPER ERROE, and ACKNOWLEDGE
 *
 * The STROBE output activates the solenoid
 * The AUTOFEED output activate the first lamp
 * The INIT output activate the second lamp
 *
 * The SELECT output is used as a HIGH reference, for the pull up resistors
 *
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <conio.h>
#include <dos.h>
#include <string.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include "jukebox.h"
#ifdef __DJGPP__
#include <go32.h>
#endif

/* LPT port, port addresses and status bits */
WORD lpt_port;
#define PORT_DTR        (lpt_port+0)

#define PORT_STR        (lpt_port+1)
#define STR_ERR         0x08
#define STR_SLCT        0x10
#define STR_PE          0x20
#define STR_ACK         0x40
#define STR_BUSY        0x80
#define STR_INPUT_MASK  (STR_ERR|STR_SLCT|STR_PE|STR_ACK|STR_BUSY)
#define STR_INVERT      (STR_ERR|STR_SLCT|STR_PE|STR_ACK)

#define PORT_CTR (lpt_port+2)
#define CTR_STOBE	0x01
#define CTR_AFD		0x02
#define CTR_INIT	0x04
#define CTR_SELECT	0x08

/* This bit needs to be high so that the output is low, otherwise
 * It interferes with the LCD
 */
#define CTR_CONFIGURE	CTR_SELECT

// Find the parallel port and set the default outputs
void switch_init(void)
{
#ifdef __DJGPP__
   lpt_port=_farpeekw(_go32_conventional_mem_selector(),0x408);
#else
  WORD far *fp;
  /* Find first LPT port and configure it with NO IRQs and required I/Os */
  *((DWORD *)(&fp))=0x00400008; /* Point to the LPT port BIOS data area */
  lpt_port= *fp;
#endif

//  screen_xy(0,14);
//  printf("Using the LPT port at 0x%X\n",lpt_port);
  if (lpt_port==0) return;
  outp(PORT_CTR,CTR_CONFIGURE);

  // Turn the relays off
  solenoid(0);
  lamp1(0);
  lamp2(0);

  // Set outputs high to start with
  outp(PORT_DTR,0xFF);
}

/* Read the PORT 80 card for at least 10uS		*/
/* To allow the capacitance on the signals		*/
static void io_delay(void)
{ int i;
  // Doesn't seem to be needed anymore !
//  for (i=0 ; i<8 ; i++) inp(0x80);
}

/* Pinouts                      */
/* D0-D7 = 2-9                  */
/* STB   = 1,   AFD  = 14       */
/* INIT  = 16                   */
/*                              */
/* ERR   = 15,  SLCT = 13       */
/* PE    = 12,  ACK  = 10       */
/* BUSY  = 11                   */
/*                              */
/* Original Wiring              */
/* Data  0 1 2 3 4 5 6 7        */
/* ERR   A B C D E F G H        */
/* SLCT  I J K L M N O P        */
/* PE    Q R S T U V 0 1        */
/* ACK   2 3 4 5 6 7 8 9        */
/*                              */
/* STB = Low -> button release  */

/* Letters buttons are bits 0-21 */
#define NUM_LETTERS 22

static char *strbin(DWORD val,int req)
{ static char retbuf[33];
  char tmpbuf[33]; 
  int len;

  val&=(1UL<<req)-1UL;

  ltoa(val,tmpbuf,2);
  len=strlen(tmpbuf);
  if (len>req)
   { strcpy(retbuf,tmpbuf);
   } else
   { memset(retbuf,'0',req-len);
     strcpy(retbuf+req-len,tmpbuf);
   }
  return retbuf;
}

/* The actual positions of the buttons A-V and 0-9 */
static BYTE key_table[32]=
 {  8, 0, 1, 9,10, 2, 3,11,     /* A-H */
   12, 4, 5,13,14, 6, 7,23,     /* J-R (No I or O) */
   16,24,25,17,30,18,26,19,     /* S-V 0-3 */
   27,20,28,21,29,22,15,31      /* 4-9 +2 spare */
 };

/* work out which buttons are currently pressed down */
static DWORD actual_switch_read(void)
{ BYTE col,in1,in2,buttons[4]; DWORD actual,trans;

  /* Nothing found so far */
  buttons[0]=buttons[1]=buttons[2]=buttons[3]=0;

  outp(PORT_DTR,0xFF); io_delay();

//  screen_xy(0,15);
  /* If they change then something was connected */
  for (col=1 ; (col&0xFF)!=0 ; col<<=1)
   { outp(PORT_DTR,~col); io_delay();
     in1=(inp(PORT_STR)^STR_INVERT)&STR_INPUT_MASK;
//printf("%s =",strbin(col,8));
//printf("%s\n",strbin(in1^in2,8));
     if (in1!=0)
      { if (in1&STR_ERR)  buttons[3]|=col;
        if (in1&STR_SLCT) buttons[2]|=col;
        if (in1&STR_ACK)  buttons[1]|=col;
        if (in1&STR_PE)   buttons[0]|=col;
      }
   }

  /* Leave the output lines high */
  outp(PORT_DTR,0xFF);


  /* Convert the acual buttons to translated buttons */
  actual= *(DWORD *)(buttons); trans=0UL;
  for (col=0 ; col<32 ; col++)
   if (actual&(1UL<<key_table[col]))
    trans|=(1UL<<col);

//printf("%s ",strbin(actual,32));
//printf("%s\n",strbin(trans,32));

  /* Return the translated buttons at a DWORD */
  return trans;
}


#define NUM_AVERAGE 5
// Read the buttons several times to debounce them
DWORD switch_read(void)
{ static DWORD last_but[NUM_AVERAGE]; static BYTE last=0;
  DWORD ored; BYTE i;

  last_but[last++]=actual_switch_read();
  last%=NUM_AVERAGE;

  ored=0UL;
  for (i=0 ; i<NUM_AVERAGE ; i++)
   ored|=last_but[i];

  return ored;
}

// Set or clear a control port bit
inline control_port_bit(int on,int bitval)
{
  // If port invalid then ignore
  if (lpt_port==0) return;

  // Clearing the control reg bit, takes the output high
  if (on)
   { outp(PORT_CTR,inp(PORT_CTR)& ~bitval);
   } else
   { outp(PORT_CTR,inp(PORT_CTR)|bitval);
   }
}

/* Enable the solenoid */
void solenoid(int on)
{ control_port_bit(on,CTR_STOBE);
}

/* Enable the first lamp */
void lamp1(int on)
{ control_port_bit(on,CTR_AFD);
}

/* Enable the second lamp */
void lamp2(int on)
{ control_port_bit(!on,CTR_INIT);
}
