/* 
 * LCD interface
 * -------------
 * copyright (c) 1998-2000 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Control an LCD display connected to the parallel port
 * uses some routines and variables from hwint.c
 *
 * It defines the access routines for 4bit mode and 8bit mode
 * I've only EVER tested the 4bit ones !
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <stdio.h>
#include <string.h>
#include "jukebox.h"
#include "font.h"

#define LCD_4BIT
/* LCD module interface file
 *
 * Pin assignment on the LCD module
 *  1 Vss  = GND
 *  2 Vcc  = +5V
 *  3 Vee  = 0.7-5V	LCD drive voltage, depends on temperature)
 *  4 RS   = H/L	H=Character, L=Command
 *  5 R/~W = H/L	H=Read, L=Write
 *  6 En   = H/L	L->H = latch RS & R/W, H->L = latch data
 *  7 D0
 *  8 D1
 *  9 D2
 * 10 D3
 * 11 D4   (D0 in 4 bit mode, high nibble first)
 * 12 D5   (D1 in 4 bit mode, high nibble first)
 * 13 D6   (D2 in 4 bit mode, high nibble first)
 * 14 D7   (D3 in 4 bit mode, high nibble first)
 */

// OR this with the byte value to send to set the bit
#define LCD_RS 0x100
#define RD_STRB 8

WORD lcd_port;

static void lcd_wait(unsigned int uS)
{ while (uS>0)
  { // 8 ISA bus reads takes about 1uS
    inp(0x80); inp(0x80); inp(0x80); inp(0x80); 
    inp(0x80); inp(0x80); inp(0x80); inp(0x80); 
    uS--;
  }
}

#ifdef LCD_4BIT
/* Wiring for 4 - bit mode
 * D0 = pin 11 (D4)
 * D1 = pin 12 (D5)
 * D2 = pin 13 (D6)
 * D3 = pin 14 (D7)
 * D7 = pin 4  (RS)
 * AF = pin 6  (En)
 * Gnd= pin 5  (R/~W)
 */

// Output the value to the parallel port untranslated
static void lcd_write_raw(WORD val)
{ BYTE cntrl;
  cntrl=inp(lcd_port+2)&(~RD_STRB);	// Clearing bit 1 puts AF output high
  outp(lcd_port,((val>>4)&0x0F)|((val>>1)&0x80));	// 4 data bits and RS bit
  lcd_wait(2);
  outp(lcd_port+2,cntrl|RD_STRB);	// Set bit - Take output low (latches DATA)
  lcd_wait(2);
  outp(lcd_port+2,cntrl);		// Clear bit - Take output high (latches RS bit)
  lcd_wait(2);
  outp(lcd_port+2,cntrl|RD_STRB);	// Set bit - Leave output low
  lcd_wait(3);
}

// Output the value as two nibbles
static void lcd_write_val(WORD val)
{ WORD rs_bit;
  rs_bit=(val&LCD_RS);
  lcd_write_raw((val&0xF0)|rs_bit);             // Send MSN and RS bit
  lcd_write_raw(((val<<4)&0xF0)|rs_bit);	// Send LSN and RS bit
}

// Reset the LCD an put it into 4 bit control mode
static void lcd_reset(void)
{ lcd_wait(10000);
  lcd_write_raw(0x30);
  lcd_wait(4100);
  lcd_write_raw(0x30);
  lcd_wait(100);
  lcd_write_raw(0x30);
  lcd_wait(4100);
  lcd_write_raw(0x20);	// Bit 4 cleared = 4 bit mode
  lcd_wait(40);
  lcd_write_val(0x28);
  lcd_wait(40);
  lcd_write_val(0x80);
  lcd_wait(40);
  lcd_write_val(0xC0);
  lcd_wait(40);
  lcd_write_val(0x60);
}
#else

// I've not tested the 8 bit code
#error The 8 bit code is untested, use at your own risk !

/* Wiring for 8 bit mode
 * D0 = pin  7 (D0)
 * D1 = pin  8 (D1)
 * D2 = pin  9 (D2)
 * D3 = pin 10 (D3)
 * D4 = pin 11 (D4)
 * D5 = pin 12 (D5)
 * D6 = pin 13 (D6)
 * D7 = pin 14 (D7)
 * ST = pin 4  (RS)
 * AF = pin 6  (En)
 * Gnd= pin 5  (R/~W)
 */

// Output the value to the parallel port
static void lcd_write_val(WORD val)
{ BYTE cntrl;
  cntrl=inp(lcd_port+2)&(~2);	// Clearing bit 1 puts AF output high
  if ((val&LCD_RS)!=0) cntrl&=(~1); // If RS is high clear ST bit (puts output high)
  outp(lcd_port,val&0xFF);		// 8 data bits
  lcd_wait(1);
  outp(lcd_port+2,cntrl|2);		// Set bit - Take output low (latches DATA)
  lcd_wait(1);
  outp(lcd_port+2,cntrl);		// Clear bit - Take output high (latches RS bit)
  lcd_wait(1);
  outp(lcd_port+2,cntrl|2);		// Set bit - Take output low (latches DATA)
}

// Reset the LCD an put it into 8 bit control mode
static void lcd_reset(void)
{ lcd_wait(15000);
  lcd_write_val(0x30);
  lcd_wait(4100);
  lcd_write_val(0x30);
  lcd_wait(100);
  lcd_write_val(0x30);
  lcd_wait(4100);
  lcd_write_val(0x30); // Bit 4 = 1 for 8 bit mode
  lcd_wait(40);
  lcd_write_val(0x38);
  lcd_wait(40);
  lcd_write_val(0x80);
  lcd_wait(40);
  lcd_write_val(0xC0);
  lcd_wait(40);
  lcd_write_val(0x60);
}
#endif

// Clear the display
static void lcd_clear(void)
{ lcd_write_val(0x01);
  lcd_wait(1650);
}

// Move cursor to top left
static void lcd_home(void)
{ lcd_write_val(0x02);
  lcd_wait(1650);
}

// move displayed characters
static void lcd_entry(char backwards,char shift)
{ lcd_write_val(0x04|(backwards ? 0 : 2)|(shift ? 1 : 0));
  lcd_wait(50);
}

// Select the display and cursor attributes
static void lcd_onoff(char disp,char cursor,char curblink)
{ lcd_write_val(0x08|(disp ? 4 : 0)|(cursor ? 2 : 0)|(curblink ? 1 : 0));
  lcd_wait(50);
}

// Scroll the display
static void lcd_shift(char display,char left)
{ lcd_write_val(0x10|(display ? 8 : 0)|(left ? 4 : 0));
  lcd_wait(50);
}

// Point to a position in the Character Graphics (User Definied Character) RAM
static void lcd_set_cg_ram(int offset)
{ lcd_write_val(0x40|(offset&0x3F));
  lcd_wait(50);
}

// Point to a position in display data RAM
static void lcd_set_dd_ram(int offset)
{ lcd_write_val(0x80|(offset&0x7F));
  lcd_wait(50);
}

// Write a value to either the CG or DD RAM
static void lcd_write_char(char ch)
{ lcd_write_val(LCD_RS|ch);
  lcd_wait(50);
}

// Move the cursor to a position on the display
static void lcd_move(int col,int row)
{ lcd_set_dd_ram(col+(row*64));
}

// Write a string to the LCD
static void lcd_write_str(char *str)
{ while (*str) lcd_write_char(*(str++));
}

// The top column number
#define LCD_OFFSET 1
#define LCD_STAT_OFF 6

static bool lcd_just_reset;

// Initialise the LCD module and write the default text to it
void lcd_init(void)
{ int i;
  extern WORD lpt_port;
  lcd_port=lpt_port;
  lcd_reset();
  lcd_clear();
  lcd_onoff(1,0,0); // Display on, cursor off, blinking cursor off

  // Clear the CG RAM
  lcd_set_cg_ram(0);
  for (i=0 ; i<64 ; i++) lcd_write_char(0);

  // Setup the 8 User defined characters at the top (left) of the display
  lcd_home();
  lcd_move(LCD_OFFSET+0,1); lcd_write_char(0);
  lcd_move(LCD_OFFSET+1,1); lcd_write_char(1);
  lcd_move(LCD_OFFSET+0,0); lcd_write_char(2);
  lcd_move(LCD_OFFSET+1,0); lcd_write_char(3);

  lcd_move(LCD_OFFSET+2,1); lcd_write_char(4);
  lcd_move(LCD_OFFSET+3,1); lcd_write_char(5);
  lcd_move(LCD_OFFSET+2,0); lcd_write_char(6);
  lcd_move(LCD_OFFSET+3,0); lcd_write_char(7);

  lcd_just_reset=true;
  lcd_update();
}

// Output a new user defined character definition
static void lcd_cg_set(char ch)
{ int i; char *cp; BYTE *chdat;
  // Force to capitals
  ch=toupper(ch);
  // find it in the lookup table
  cp=strchr(chrs,ch);
  // If not found set to last character (SPACE)
  if (cp==NULL) cp=chrs+strlen(chrs)-1;
  // Copy the data from the font table
  for (i=0,chdat=font+16*(cp-chrs) ; i<16 ; i++,chdat++)
   lcd_write_char(*chdat);
}

// Write a string at a given location
static void lcd_write_string(int col,int row,char *str)
{ lcd_move(col,row);
  while (*str) lcd_write_char(*(str++));
}

// Output a string to a selected position
// 0=Top 2 digit string, 1=Bottom 2 digit string
// 2=Left string, 3=Right string
static void lcd_display(int pos,char *str)
{ if (pos==0 || pos==1)
  { lcd_set_cg_ram(pos*32);
    lcd_cg_set(str[0]);
    lcd_cg_set(str[1]);
  } else
  { lcd_write_string(LCD_OFFSET+LCD_STAT_OFF,pos-2,str);
  }
}

extern trk_info_t *trks;
extern int ntrks;
extern play_list_t toplay[];
extern int ntoplay,total_played;
extern bool locked,background_mode;

// Update the LCD displayed info based on the current jukebox status
void lcd_update(void)
{ static int last_idx,last_next,last_toplay,last_total;
  static bool last_lock,last_back;
  char buf[8];
  int idx;

  if (lcd_just_reset)
  { last_idx= -1; last_next= -1; last_toplay= -1; last_total= -1;
	last_lock= -1; last_back= -1;
	lcd_just_reset=false;
  }

  idx=(ntoplay>=1) ? toplay[0].trk_index : -2;
  if (last_idx!=idx)
  { lcd_display(0,(idx>=0) ? track_label(idx) : "  ");
    last_idx=idx;
  }

  idx=(ntoplay>=2 && background_mode==false) ? toplay[1].trk_index : -2;
  if (last_next!=idx)
  { lcd_display(1,(idx>=0) ? track_label(idx) : "  ");
    last_next=idx;
  }

  if (ntoplay!=last_toplay || total_played!=last_total)
  { sprintf(buf,"%3u %2u",total_played,ntoplay); 
	lcd_display(3,buf);
    last_toplay=ntoplay;
	last_total=total_played;
  }

  if (last_lock!=locked)
  { lcd_move(LCD_OFFSET+LCD_STAT_OFF-1,0);
    lcd_write_char((locked) ? 'l' : ' ');
    last_lock=locked;
  }

  if (last_back!=background_mode)
  { lcd_move(LCD_OFFSET+LCD_STAT_OFF-1,1);
    lcd_write_char((background_mode) ? 'b' : ' ');
    last_back=background_mode;
  }
}

void lcd_update_time(unsigned seconds)
{ char buf[8];
  sprintf(buf,"%2u:%02u",seconds/60,seconds%60);
  lcd_display(2,buf);
}

// Clear the LCD ready for EXIT
void lcd_exit(void)
{ lcd_write_string(0,0,"                ");
  lcd_write_string(0,1,"                ");
  lcd_write_string(LCD_OFFSET,0,"Jukebox");
  lcd_write_string(LCD_OFFSET,1,"Stopped");
}
