/* 
 * Serial port buffering
 * ---------------------
 * copyright (c) 1987-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Provide buffered serial input and output routines
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <ctype.h>
#include <conio.h>
#include "mlib.h"

// TX & RX ports
#define COM_TX	0
#define COM_RX	0

// Interrupt enable register
#define COM_IER		1
#define COM_IER_EDSSI	0x08	// Enable Delta Status Signal Interrupt
#define COM_IER_ELSI	0x04	// Enable Delta Status Signal Interrupt
#define COM_IER_ETBEI	0x02	// Enable TX buffer empty Interrupt
#define COM_IER_ETBFI	0x01	// Enable RX buffer full Interrupt

// Interrupt identification register
#define COM_IIR 2
#define COM_IIR_FIFO2	0x80
#define COM_IIR_FIFO1	0x40
#define COM_IIR_ID1		0x04
#define COM_IIR_ID2		0x08
#define COM_IIR_ID1		0x04
#define COM_IIR_ID0		0x02
#define COM_IIR_NOPENDING	0x01

// Line Control Register
#define COM_LCR	3
#define COM_LCR_DLAB	0x80	// Divisor latch addressing bit
#define COM_LCR_SBR		0x40	// Set BReak
#define COM_LCR_SPAR	0x20	// Sticky parity
#define COM_LCR_EPAR	0x10	// Even parity
#define COM_LCR_PAREN	0x08	// Parity enabled
#define COM_LCR_SB		0x04	// Stop bit extend (to 1.5 or 2)
#define COM_LCR_BITS	0x03	// Mask for bit modes
#define COM_LCR_5BIT	0x00
#define COM_LCR_6BIT	0x01
#define COM_LCR_7BIT	0x02
#define COM_LCR_8BIT	0x03


// Modem Control Register
#define COM_MCR	4
#define COM_MCR_AFE		0x20	// Automatic flow enable (Some Texas TL16C550C's only)
#define COM_MCR_LOOP	0x10	// Local loopback
#define COM_MCR_OUT2	0x08	// If 0 IRQ out is high-impedance
#define COM_MCR_OUT1	0x04	// Set to 1 for PC usage
#define COM_MCR_RTS		0x02	// Controls RTS output pin
#define COM_MCR_DTR		0x01	// Controls DTR output pin


// Line Status Register
#define COM_LSR	5
#define COM_LSR_RXER	0x80	// At least one character in the FIFO has an error (16550+ only)
#define COM_LSR_TEMT	0x40	// No TX running, reset by clearing LSR
#define COM_LSR_THRE	0x20	// TX holding reg empty, (FIFO's = all words empty)
#define COM_LSR_BI		0x10	// Break Indicator (more than a word of "Space")
#define COM_LSR_FE		0x08	// Framing error
#define COM_LSR_PE		0x04	// Parity error
#define COM_LSR_OE		0x02	// Overrun error
#define COM_LSR_DR		0x01	// Data Ready


// Modem Status Register
#define COM_MSR	6
#define COM_MSR_DCTS	0x01	// Delta CTS, CTS changed since last read
#define COM_MSR_DDSR	0x02	// Delta DSR, DSR changed since last read
#define COM_MSR_DRI		0x04	// Delta RI, RI changed since last read
#define COM_MSR_DDCD	0x08	// Delta DCD, DCD changed since last read
#define COM_MSR_CTS		0x10	// CTS input
#define COM_MSR_DSR		0x20	// DSR input
#define COM_MSR_RI		0x40	// RI input
#define COM_MSR_DCD		0x80	// DCD input

static WORD com_port=0;
static BYTE com_irq;
static irq_info_t IrqInfo;
static BYTE com_fifo_size;
static BYTE old_ier,old_mcr;

// return the interrupt line a serial port is using
// returns: 0xFF if no interrupt found, or intlevel 0-15
static BYTE detect_IRQ(WORD base)
{ BYTE ier,mcr,imrm,imrs,maskm,masks,irqm,irqs;

// _asm cli;						// disable all CPU interrupts

  ier = inp(base+COM_IER);		// read IER
  mcr = inp(base+COM_MCR);		// read MCR

  outp(base+COM_IER,0x00);		// disable all UART ints

  // wait for the THR to be empty
  while ((inp(base+COM_LSR)&COM_LSR_THRE)!=0)
  { // Do nothing
  }  

  // connect UART to irq line
  outp(base+COM_MCR,COM_MCR_OUT2|inp(base+COM_MCR));

  imrm = inp(0x21);			// read contents of master ICU mask register
  imrs = inp(0xA1);			// read contents of slave ICU mask register
  outp(0xA0,0x0A);			// next read access to 0xA0 reads out IRR
  outp(0x20,0x0A);			// next read access to 0x20 reads out IRR

  outp(base+COM_IER,COM_IER_ETBEI);	// Enable TX buffer empty interrupts...
  maskm = inp(0x20);		// this clears all bits except for the one
  masks = inp(0xA0);		// that corresponds to the int
  outp(base+COM_IER,0);		// drop the int line

  maskm &= ~inp(0x20);		// this clears all bits except for the one
  masks &= ~inp(0xA0);		// that corresponds to the int

  outp(base+COM_IER,COM_IER_ETBEI);	// raise it again just to be sure...

  maskm &= inp(0x20);		// this clears all bits except for the one
  masks &= inp(0xA0);		// that corresponds to the int
  outp(0xA1,~masks);		// now let us unmask this interrupt only
  outp(0x21,~maskm);

  outp(0xA0,0x0C);		// enter polled mode; Mike Surikov reported
  outp(0x20,0x0C);		// that order is important with Pentium/PCI systems

  irqs = inp(0xA0);		// and accept the interrupt
  irqm = inp(0x20);
  inp(base+COM_IIR);		// reset transmitter interrupt in UART
  
  outp(base+COM_MCR,mcr);	// restore old value of MCR
  outp(base+COM_IER,ier);	// restore old value of IER
  
  if (masks) outp(0xA0,0x20);	// send an EOI to slave
  if (maskm) outp(0x20,0x20);	// send an EOI to master
  
  outp(0x21,imrm);			// restore old mask register contents  
  outp(0xA1,imrs);
//   _asm  sti;

  // Work out what happened
  if (irqs&0x80)		// slave interrupt occured
    return (irqs&0x07)+8;
  if (irqm&0x80)		// master interrupt occured
    return irqm&0x07;
  return 0xFF;
}

#define COM_BUF_SIZE 4096
static BYTE tx_buf[COM_BUF_SIZE],rx_buf[COM_BUF_SIZE];
static int  tx_in=0,tx_out=0;
static int  rx_in=0,rx_out=0;

static BYTE fifo_get(BYTE *buf,int *in,int *out)
{ int pos; BYTE val;

  // Wait for buffer to get some data
  while ( (*in)==(*out) )
  { fputs("",stderr);
    usleep(10);
  }

  pos=(*out); 
  val=buf[pos]; 
  *out=(pos+1)%COM_BUF_SIZE;  

  return val;
}

static void fifo_put(BYTE *buf,int *in,int *out,BYTE val)
{ int next,pos;

  // Wait for some space
  while ( (next=((pos=(*in))+1)%COM_BUF_SIZE)==(*out) )
  { fputs("",stderr);
    usleep(10);
  }

  *in=next;
  buf[pos]=val; 
}

static volatile int nSent=0;

static irq_interrupt_t com_irq_handler(void)
{ BYTE cause;

  cause=inp(com_port+COM_IIR);
  if ((cause&COM_IIR_NOPENDING)!=0)
  { // Not me, chain IRQ
	// call (old_vect)();
	IrqAck(&IrqInfo);
    return;
  }

  // ACK the IRQ
  IrqAck(&IrqInfo);

  // Stop more from coming
  outp(com_port+COM_MCR,old_mcr& ~COM_MCR_OUT2);

  while ((cause&COM_IIR_NOPENDING)==0)
  { BYTE stat;
    // Switch on the cause until they go away
    switch (cause & (COM_IIR_ID0|COM_IIR_ID1))
	{ case 0:	// MODEM status changed
		inp(com_port+COM_MSR);	// reading here clears the cause
		break;

	  case COM_IIR_ID1|COM_IIR_ID0:	// Status changed
		inp(com_port+COM_LSR);	// reading here clears the cause
		break;

	  case COM_IIR_ID1: // RX
		do { fifo_put(rx_buf,&rx_in,&rx_out,inp(com_port+COM_RX));
			 stat=inp(com_port+COM_LSR);
	       }
		while ((stat&COM_LSR_DR)!=0);
		if ((stat&COM_LSR_THRE)==0)
		{ // TX holding not reg empty so bail out
		  break;
		}
		// DROP INTO TX CODE, BECAUSE WE KNOW THAT THE TX BUFFER HAS SPACE
	  case COM_IIR_ID0:
		// The buffer is empty, so remember it
	    nSent=0;
		while (1)
		{ if (tx_in==tx_out)
		  { // Nothing left to send so disable IRQ
		    outp(com_port+COM_IER,inp(com_port+COM_IER)& ~COM_IER_ETBEI);
			break;
		  } else
		  if (nSent<com_fifo_size)
		  { // Space in FIFO so send next byte(s) from buffer
		    outp(com_port+COM_TX,fifo_get(tx_buf,&tx_in,&tx_out));
			nSent++;
		  } else
		  { // No more space in TX FIFO so bail out
			break;
		  }
		}
		break;
	}
	cause=inp(com_port+COM_IIR);
  }

  // Re-enable
  outp(com_port+COM_MCR,old_mcr|COM_MCR_OUT2);
}


// EXPORTED functions

int serial_init(void)
{ 
  // Already done, ignore
  if (com_port!=0) return 0;

#ifdef __DJGPP__
  com_port=_farpeekw(_go32_conventional_mem_selector(),0x400);
#else
  { WORD far *fp;
    /* Find first COM port */
    *((DWORD *)(&fp))=0x00400000; /* Point to the COM1 port BIOS data area */
    com_port= *fp;
  }
#endif
  // No com port found
  if (com_port==0) 
  {	fprintf(stderr,"No comport found\n");
	return 1;
  }

  fprintf(stderr,"Com port at 0x%X",com_port);

  if (com_port==0x3f8) com_irq=4; else
  if (com_port==0x2f8) com_irq=3; else
  com_irq=detect_IRQ(com_port);
  if (com_irq<2 || com_irq>15) 
  {	fprintf(stderr,"Com port at 0x%03X has no IRQ connected\n",com_port);
    com_port=0;
	return 2;
  }
  fprintf(stderr,", IRQ %d\n",com_irq);

  tx_in=tx_out=rx_in=rx_out=0;
  com_fifo_size=1;

  system("mode com1:9600,n,8,1") ;

  // Remember original values of interrupt mode and enable registers
  old_mcr=inp(com_port+COM_MCR);
  old_ier=inp(com_port+COM_IER);
  
  // Disable interupts at source
  outp(com_port+COM_MCR,old_mcr& ~COM_MCR_OUT2);
  outp(com_port+COM_IER,0);

  // Check for a 16550A?
  if ( (inp(com_port+COM_IIR)&(COM_IIR_FIFO2|COM_IIR_FIFO1))==(COM_IIR_FIFO2|COM_IIR_FIFO1))
  { com_fifo_size=10;
    outp(com_port+COM_IIR,0xC7); // Enable FIFOs size
  }

  if (IrqInit(&IrqInfo,com_irq,com_irq_handler))
  { fputs("Error: Serial IRQ didn't workout",stderr);
  }

  // Clear status register IRQs
  inp(com_port+COM_LSR);
  inp(com_port+COM_MSR);

  // Enable interupts at source
  outp(com_port+COM_MCR,old_mcr|COM_MCR_OUT2);

  // Only enable RX interrupts until there is something to send
  outp(com_port+COM_IER,inp(com_port+COM_IER)|COM_IER_ETBFI);

  // All OK
  return 0;
}

void serial_exit(void)
{ if (com_port==0) return;

  // Disable interupts
  outp(com_port+COM_MCR,old_mcr& ~COM_MCR_OUT2);
  outp(com_port+COM_IER,0);
  IrqEnd(&IrqInfo);

  if (com_fifo_size!=1)
  { com_fifo_size=1;
	outp(com_port+COM_IIR,0x00); // Disable FIFOs
  }

  // Restore old interrupt mode and enable flags
  outp(com_port+COM_IER,old_ier);
  outp(com_port+COM_IER,old_mcr);

  com_port=0;
}

void serial_tx_clear(void)
{ outp(com_port+COM_IER,inp(com_port+COM_IER)& ~COM_IER_ETBEI);
  tx_in=tx_out=0;
}

BOOL serial_tx_empty(void)
{ return (tx_in==tx_out && (inp(com_port+COM_LSR)&COM_LSR_THRE)!=0);
}

void serial_tx(BYTE ch)
{ 
  if (tx_in==tx_out && (inp(com_port+COM_LSR)&COM_LSR_THRE)!=0)
  { outp(com_port+COM_TX,ch);
  } else
  { // Add character to our software TX-FIFO
    fifo_put(tx_buf,&tx_in,&tx_out,ch);
  }
  // FIFO was empty, re-enable IRQs
  outp(com_port+COM_IER,inp(com_port+COM_IER)|COM_IER_ETBEI);
}

// Check to see if the serial receive buffer has some data
BOOL serial_rx_ready(void)
{ return (rx_in!=rx_out);
}

BYTE serial_rx(void)
{ if (com_port==0) return 0x00;
  return fifo_get(rx_buf,&rx_in,&rx_out);
}


#ifdef TEST_PROG
static void dump_bin(char *name,int offset)
{ BYTE val,mask;
  val=inp(com_port+offset);
  printf("Reg %s (0x%02X) = %02X = ",name,offset,val);
  for (mask=0x80 ; mask!=0 ; mask>>=1)
  { printf("%d",(val&mask)!=0 ? 1 : 0);
  }
  printf("\n");
}

int main(int argc,char *argv[])
{ int i,ch;
  serial_init();

  do { serial_buffered_tx("abcDEFGHIJKLMNOPQRSTUVWXY");
       for (i=1 ; i<80 ; i++)
	   { serial_buffered_tx("ABCDEFGHIJKLMNOPQRSTUVWXY");
	   }
	   while (tx_buf_pos!=tx_buf_len)
	   { serial_buffered_poll();
	     if (serial_rx_ready())
		 { fputc(serial_rx(),stderr);
		 }
	   }
	   serial_tx('@');

	   ch=0;
	   if (kbhit())
	   { ch=getch(); if (ch==0) ch=256+getch();
	     if (ch==' ')
		 { printf("\n");
		   dump_bin("IER",COM_IER);
		   dump_bin("IIR",COM_IIR);
		   dump_bin("LCR",COM_LCR);
		   dump_bin("MCR",COM_MCR);
		   dump_bin("LSR",COM_LSR);
		   dump_bin("MSR",COM_MSR);
		 }
             if (ch=='b')
             { int i;
               outp(com_port+COM_LCR,inp(com_port+COM_LCR)|COM_LCR_SBR);
		ch=getch(); if (ch==0) ch=256+getch();
               outp(com_port+COM_LCR,inp(com_port+COM_LCR)& ~COM_LCR_SBR);
             }
	   }
	 }
  while (ch!=27);
  serial_exit();
}
#endif
